function numex(num)
%%% project: Comparing (Empirical-Gramian-Based) Model Order Reduction Algorithms
%%% version: 1.0 (2020-02-25)
%%% authors: C. Himpe (0000-0003-2194-6754)
%%% license: BSD 2-Clause License (opensource.org/licenses/BSD-2-Clause)
%%% summary: Companion code computing numerical results

% input:  *num* sets the parameter-space dimension: 0 (Default), 1, or 4.
% output: CSV files named morscore*num*-{1,2,8}.csv storing computed MORscores.

    global ODE;

    warning off all;

    load('ABCE.mat'); % Loads A0, A1, A2, A3, A4, B, C, E

    fom = struct('A0',A0,'A1',A1,'A2',A2,'A3',A3,'A4',A4,'B',B,'C',C,'E',E);	% FOM structure specific for 4-component affine system

    pmin = 1e0;									% Minimum parameter (component) value
    pmax = 1e1;									% Maximum parameter (component) value
    pnom = 10.^mean([log10(pmin),log10(pmax)]);					% Nominal parameter (component) value

    ntrain = 3;									% Number of training parameters
    ntest = 10;									% Number of testing parameters

    % fom.gen is the parametric system matric generator
    % ptrain is the parameter sampling for the offline phase training
    % ptest is the parameter sampling for the online phase testing
    switch(num)

        case 1

            fom.gen = @(A0,A1,A2,A3,A4) @(p) A0 + (0.2*p)*A1 + (0.4*p)*A2 + (0.6*p)*A3 + (0.8*p)*A4;
            ptrain = 10.^linspace(log10(pmin),log10(pmax),ntrain);
            ptest = 10.^(log10(pmin) + (log10(pmax)-log10(pmin)) * rand(1,ntest));

        case 4

            fom.gen = @(A0,A1,A2,A3,A4) @(p) A0 + p(1)*A1 + p(2)*A2 + p(3)*A3 + p(4)*A4;
            ptrain = 10.^linspace(log10(pmin),log10(pmax),ntrain) - pnom;
            ptrain = blkdiag(ptrain,ptrain,ptrain,ptrain) + pnom;
            ptest = 10.^(log10(pmin) + (log10(pmax)-log10(pmin)) * rand(4,ntest));

        otherwise

            fom.gen = @(A0,A1,A2,A3,A4) @(p) A0 + (0.2*p)*A1 + (0.4*p)*A2 + (0.6*p)*A3 + (0.8*p)*A4;
            ptrain = pnom;
            ptest = ptrain;
    end

    fom.A = fom.gen(A0,A1,A2,A3,A4);						% Instantiate FOM system matrix generator

    f = @(x,u,p,t) 0;								% Primal vector field
    F = @(x,u,p,t) 1;								% Adjoint vector field

    M = size(B,2);								% Number of inputs
    N = size(B,1);								% Number of states
    Q = size(C,1);								% Number of outputs

    dt = 0.01;									% Time step width
    Tf = 4.0;									% Training time horizon (needs to be large enough to reach steady state)
    Tt = 1.0;									% Testing time horizon

    ODE = @(f,g,t,x0,u,p) rk1i(fom.A,fom.B,fom.C,fom.E,t,x0,u,p,f(0,0,0,0));	% set emgr ODE solver for emgr

    % These lines plot the time-domain output trajectory and quit:
    %X = rk1i(fom.A,fom.B,fom.C,fom.E,[dt,Tf],zeros(N,1),@(t) (t<=dt)./dt,1e1);
    %figure; plot(0:dt:Tf,X); return;

    %% Compute empirical Gramians and their SVDs up to order 100.
    nf = [3,0,0,0,0,0,0,0,0,0,0,0];
    ut = 'i';

    max_order = 100;								% Maxmum rank of empirical Gramians

    ti = tic;
    WC = emgr(f,1,[M,N,Q],[dt,Tf],'c',ptrain,nf,ut);
    [UC,SC,~] = svds(WC,max_order); SC = diag(SC);
    fprintf('Computed Empirical Controllability Gramian WC (%g s) \n',toc(ti));

    ti = tic;
    WO = emgr(F,1,[Q,N,M],[dt,Tf],'c',ptrain,nf,ut);
    [UO,SO,~] = svds(WO,max_order); SO = diag(SO);
    fprintf('Computed Empirical Observability Gramian WO (%g s) \n',toc(ti));

    ti = tic;
    WX = emgr(f,F,[M,N,Q],[dt,Tf],'y',ptrain,nf + [0,0,0,0,0,0,1,0,0,0,0,0],ut);
    [UX,SX,VX] = svds(WX,max_order); SX = diag(SX);
    fprintf('Computed Empirical Linear Non-Symmetric Cross Gramian WZ (%g s) \n',toc(ti));

    fprintf('\n');


    %% Define Reductors
    reductors = {};								% Initialize reductor container


    %% Empirical Poor Man

    % WC
    reductors(end+1,1:2) = {@(n,p) makerom(fom,UC,UC,n),'Empirical Poor Man (WC)'};
    fprintf('Prepared %s \n', reductors{end,2});

    % WO
    reductors(end+1,1:2) = {@(n,p) makerom(fom,UO,UO,n),'Empirical Poor Man (WO)'};
    fprintf('Prepared %s \n', reductors{end,2});


    %% Empirical Approximate Balancing
    
    % WC, WO
    reductors(end+1,1:2) = {@(n,p) makerom(fom,UO,UC,n),'Empirical Approximate Balancing (WC,WO)'};
    fprintf('Prepared %s \n', reductors{end,2});

    % WX
    reductors(end+1,1:2) = {@(n,p) makerom(fom,VX,UX,n),'Empirical Approximate Balancing (WX)'};
    fprintf('Prepared %s \n', reductors{end,2});


    %% Empirical Dominant Subspaces

    % WC, WO
    [DSco,~,~] = svd([(UC.*SC')./norm(SC),(UO.*SO')./norm(SO)],'econ');
    DSco = DSco(:,1:min(max_order,size(DSco,2)));

    reductors(end+1,1:2) = {@(n,p) makerom(fom,DSco,DSco,n),'Empirical Dominant Subspaces (WC,WO)'};
    fprintf('Prepared %s \n', reductors{end,2});

    % WX
    [DSwx,~,~] = svd([UX.*SX',VX.*SX'],'econ');
    DSwx = DSwx(:,1:min(max_order,size(DSwx,2)));

    reductors(end+1,1:2) = {@(n,p) makerom(fom,DSwx,DSwx,n),'Empirical Dominant Subspaces (WX)'};
    fprintf('Prepared %s \n', reductors{end,2});


    %% Empirical Balanced Truncation

    % WC, WO
    LC = UC .* sqrt(SC)';
    LO = UO .* sqrt(SO)';
    [U,HSVco,V] = svd(LO'*LC,'econ');
    HSVco = diag(HSVco);
    y = sqrt(HSVco)';
    BTcoL = LO * (U ./ y);
    BTcoR = LC * (V ./ y);

    reductors(end+1,1:2) = {@(n,p) makerom(fom,BTcoL,BTcoR,n),'Empirical Balanced Truncation (WC,WO)'};
    fprintf('Prepared %s \n', reductors{end,2});

    % WX
    [LO,EO] = eigs(WX,max_order); EO = diag(EO);
    [LC,EC] = eigs(WX',max_order); EC = diag(EC);

    LC = LC .* sqrt(abs(EC))';
    LO = LO .* sqrt(abs(EO))';
    [U,HSVwx,V] = svd(LC'*LO,'econ');
    HSVwx = diag(HSVwx);
    y = sqrt(HSVwx)';
    BTwxL = LC * (U ./ y);
    BTwxR = LO * (V ./ y);

    reductors(end+1,1:2) = {@(n,p) makerom(fom,BTwxL,BTwxR,n),'Empirical Balanced Truncation (WX)'};
    fprintf('Prepared %s \n', reductors{end,2});


    %% Empirical Balanced Gains

    % WC, WO
    DIco = sum((BTcoL' * fom.B).^2 .* HSVco,2);
    [~,IX] = sort(DIco,'descend');
    BGcoL = BTcoL(:,IX);
    BGcoR = BTcoR(:,IX);

    reductors(end+1,1:2) = {@(n,p) makerom(fom,BGcoL,BGcoR,n),'Empirical Balanced Gains (WC,WO)'};
    fprintf('Prepared %s \n', reductors{end,2});

    % WX
    DIwx = sum((BTwxL' * fom.B).^2 .* HSVwx,2);
    [~,IX] = sort(DIwx,'descend');
    BGwxL = BTwxL(:,IX);
    BGwxR = BTwxR(:,IX);

    reductors(end+1,1:2) = {@(n,p) makerom(fom,BGwxL,BGwxR,n),'Empirical Balanced Gains (WX)'};
    fprintf('Prepared %s \n', reductors{end,2});

    fprintf('\n');


    %% Define Norms

    % Helper variables
    CC = fom.C  * UC;
    BO = fom.B' * UO;
    CS = UC .* SC'; 
    CQ = UC .* sqrt(SC)';
    OS = UO .* SO';
    OQ = UO .* sqrt(SO)';

    norms = {};									% Initialize norms container holding {FOM norm, ROM norm, norm name}

    norms(end+1,1:3) = {@(y)   sum(abs(prod(y,1).^(1/size(y,1)))), ...
                        @(y,m) sum(abs(prod(y,1).^(1/size(y,1)))),'Lebesgue-0'};% Approximate Discrete L_0 Signal "Norm"

    norms(end+1,1:3) = {@(y)   dt * norm(y(:),1), ...
                        @(y,m) dt * norm(y(:),1),'Lebesgue-1'};			% Approximate Discrete L_1 Signal Norm

    norms(end+1,1:3) = {@(y)   sqrt(dt) * norm(y(:),2), ...
                        @(y,m) sqrt(dt) * norm(y(:),2),'Lebesgue-2'};		% Approximate Discrete L_2 Signal Norm

    norms(end+1,1:3) = {@(y)   norm(y(:),Inf), ...
                        @(y,m) norm(y(:),Inf),'Lebesgue-\infty'};		% Approximate Discrete L_infinity Signal Norm

    norms(end+1,1:3) = {@(y)   sqrt(abs(trace( (CC.*SC')*CC' ))), ...
                        @(y,m) sqrt(abs(trace( (CC.*SC')*CC' - (m.C*m.L*CS)*CC' - (CC.*SC')*(UC'*m.R*m.C') + (m.C*m.L*CS)*(UC'*m.R*m.C') ))),'Hardy-2'};	% Approximte H_2 Norm

    norms(end+1,1:3) = {@(y)   2.0 * sum(HSVco), ...
                        @(y,m) 2.0 * sum(svd( (OQ - m.R*(m.L*OQ))'*(CQ - m.R*(m.L*CQ)) )),'Hardy-\infty'};	% Approximate H_infinity Norm

    norms(end+1,1:3) = {@(y)   sqrt(sum(HSVco.^2)), ...
                        @(y,m) sqrt(sum(svd( (OQ - m.R*(m.L*OQ))'*(CQ - m.R*(m.L*CQ)) ).^2)),'HSH'};		% Approximate Hilbert-Schmidt-Hankel Norm

    norms(end+1,1:3) = {@(y)   HSVco(1), ...
                        @(y,m) svds( (OQ - m.R*(m.L*OQ))'*(CQ - m.R*(m.L*CQ)) ,1),'Hankel'}; 			% Approximate Hankel Norm

    norms(end+1,1:3) = {@(y)   sqrt(abs(eigs( (CC.*SC')*CC',1))), ...
                        @(y,m) sqrt(abs(eigs( (CC.*SC')*CC' - (m.C*m.L*CS)*CC' - (CC.*SC')*(UC'*m.R*m.C') + (m.C*m.L*CS)*(UC'*m.R*m.C'), 1))),'Primal Modified Hankel'};	% Approximate Hankel Induced Primal Norm 

    norms(end+1,1:3) = {@(y)   sqrt(abs(eigs( (BO.*SO')*BO',1))), ...
                        @(y,m) sqrt(abs(eigs( (BO.*SO')*BO' - (m.B'*m.L*OS)*BO' - (BO.*SO')*(UO'*m.R*m.B) + (m.B'*m.L*OS)*(UO'*m.R*m.B), 1))),'Dual Modified Hankel'};		% Approximate Hankel Induced Dual Norm

    %% Test ROMs
    max_order = 50;								% Maximum reduced order tested
    x0 = zeros(N,1);

    rand('seed',1009);
    ur = rand(1,floor(Tt/dt)+1);
    u = @(t) ur(1,floor(t/dt)+1);						% Use random input function

    params = num2cell(ptest,1);
    err_rom = ones(numel(params),numel(reductors(:,1)),numel(norms(:,1))+1,max_order);
    sta_rom = zeros(numel(params),numel(reductors(:,1)));

    fprintf('Testing %u methods for %u parameter samples in %u norms \n',numel(reductors(:,1)),numel(params),numel(norms(:,1)));

    % For each parameter sample ...
    for k = 1:numel(params)

        y = rk1i(fom.A,fom.B,fom.C,fom.E,[dt,Tt],x0,u,params{k});

        % For each norm compute FOM norm
        for n = 1:numel(norms(:,1))

            n_fom{n} = norms{n,1}(y);
        end%for

        % For each reductor ...
        for l = 1:numel(reductors(:,1))

            % For up to max order compute reduced output trajectory
            for m = 1:max_order

                rom = reductors{l,1}(m,params{k});

                yr = rk1i(rom.A,rom.B,rom.C,rom.E,[dt,Tt],zeros(m,1),u,params{k}); % Note zero initial conditions are assumed

                % For each norm compute bounded norm
                for n = 1:numel(norms(:,1))

                    err_rom(k,l,n,m) = max(1e-16,min(1.0,norms{n,2}(y - yr,rom)/n_fom{n}));
                end%for

                % Check stability
                try

                    sta_rom(k,l) = sta_rom(k,l) + (real(eigs(rom.A(params{k}),rom.E,1,'lr')) >= 0);
                catch

                    sta_rom(k,l) = NaN;
                end%try
            end%for
        end%for

        fprintf('#');
    end%for

    fprintf('\n');

    % Apply parametric norms
    for l = 1:numel(reductors(:,1))

        for n = 1:numel(norms(:,1))

            ms1{l,n} = morscore(1:max_order,mean(err_rom(:,l,n,:),1));
            ms2{l,n} = morscore(1:max_order,sqrt(sum(err_rom(:,l,n,:).^2,1)));
            ms8{l,n} = morscore(1:max_order,max(err_rom(:,l,n,:),[],1));
        end%for

        ms1{l,end+1} = mean(sta_rom(:,l));
        ms2{l,end+1} = sqrt(sum(sta_rom(:,l).^2));
        ms8{l,end+1} = max(sta_rom(:,l));
    end%for    

    % Save CSV files and output MORscores
    if not(num == 0)

        dlmwrite(['morscore',num2str(num),'-1.csv'],cell2mat(ms1),'precision','%.2f');
        dlmwrite(['morscore',num2str(num),'-2.csv'],cell2mat(ms2),'precision','%.2f');
        dlmwrite(['morscore',num2str(num),'-8.csv'],cell2mat(ms8),'precision','%.2f');

        MORscore_1 = cell2mat(ms1)
        AVG_1 = mean(MORscore_1(:,1:end-1),2)
        MORscore_2 = cell2mat(ms2)
        AVG_2 = mean(MORscore_2(:,1:end-1),2)
        MORscore_8 = cell2mat(ms8)
        AVG_8 = mean(MORscore_8(:,1:end-1),2)
    else

        dlmwrite(['morscore',num2str(num),'.csv'],cell2mat(ms8),'precision','%.2f');

        MORscore = cell2mat(ms2)
        AVG = mean(MORscore(:,1:end-1),2)
    end%if

    % Classic error plots
    err_rom(err_rom==1) = NaN;

    cmap = lines12(10);

    for n = 1:numel(norms(:,1))

        figure('NumberTitle','off','Name',norms{n,3});
        semilogy(1:max_order,squeeze(err_rom(1,1,n,:)),'linewidth',3,'Color',cmap(1,:));

        hold on;

        for l = 2:numel(reductors(:,1))

            semilogy(1:max_order,squeeze(err_rom(1,l,n,:)),'linewidth',3,'Color',cmap(l,:));
        end%for

        hold off;

        legend(reductors(:,2),'Location','SouthOutside');
        ylim([1e-16,1]);
        xlim([1,max_order]);
    end%for
end

%% makerom
function s = makerom(fom,L,R,n)
% summary: Assembles a ROM from a FOM structure.
% input:   *fom* is the full order model structure.
%          *L* is the full left projector.
%          *R* is the full right projector.
%          *n* is the target reduced order (a positive integer).
% output:  *r* is a matrix of dimension (n times 3) holding n RGB colors. 

    s.L = L(:,1:n)';
    s.R = R(:,1:n);
    s.A = fom.gen(s.L*fom.A0*s.R,s.L*fom.A1*s.R,s.L*fom.A2*s.R,s.L*fom.A3*s.R,s.L*fom.A4*s.R);
    s.B = s.L*fom.B;
    s.C = fom.C*s.R;
    s.E = s.L*fom.E*s.R;
end

%% morscore
function s = morscore(orders,errors)
% summary: Computes the MORscore.
% input:   *orders* is a vector of tested reduced orders.
%          *errors* is a vector of model reduction errors to the orders.
% output:  *s* is the computed MORscore. 

    nx = orders./max(orders);
    ny = log10(errors)./floor(log10(eps));

    if not(nx(end)==1),

        nx(end+1) = 1;
        ny(end+1) = ny(end);
    end%if

    s = max(0,trapz(nx(:),ny(:)));
end

%% rk1i
function y = rk1i(A,B,C,E,t,x0,u,p,s)
% summary: Implicit Euler Integrator (with decomposition caching).
% input:   *A* is function handle to a parametric system matrix generator.
%          *B* is an input matrix.
%          *C* is an output matrix.
%          *E* is a mass matrix.
%          *t* is a two component vector holding time-step and time horizon.
%          *x0* is the initial condition vector.
%          *u* is a function handle to the input function.
%          *p* is the parameter vector.
%          *s* is a Boolean, with 0 for the primal and 1 for the dual state
%              trajectory, if *s* is empty an output trajectory is generated.
% output:  *y* is the resulting trajectory matrix.

    persistent lea;
    persistent uea;
    persistent pea;
    persistent pr;

    n = numel(x0);
    h = t(1);
    K = floor(t(2)/h) + 1;

    % Caching the primal and dual LU decompositions
    if isempty(pr) || not(size(x0,1) == numel(pea)) || not(all(pr == p))

        [lea,uea,pea] = lu(E - h*A(p),'vector');
        pr = p;
    end%if

    x = x0;

    if nargin == 9, c = 1; else, c = C; s = 0; end%if

    if not(s) % primal snapshots

        y = c * x;
        y(:,K) = 0;

        for k = 2:K

            t = (k - 1.5) * h;
            z = E * x + h * B * u(t);

            x = uea \ (lea \ z(pea));
            y(:,k) = c * x;
        end%for    

    else % dual snapshots

        y = c * x;
        y(:,K) = 0;

        for k = 2:K

            t = (k - 1.5) * h;
            z = (x' * E)' + h * (u(t)' * C)';
        
            x = ((z' / uea) / lea)';
            y(:,k) = c * x(pea);
        end%for
    end%if
end

%% lines12
function r = lines12(n)
% summary: Color map for 12 distinguishable line colors.
% input:   *n* is the number of requested colors.
% output:  *r* is a matrix of dimension (n times 3) holding n RGB colors.

    if nargin<1 || isempty(n), n = 10; end%if

    x = [166,206,227; ...
         31,120,180;  ...
         178,223,138; ...
         51,160,44;   ...
         251,154,153; ...
         227,26,28;   ...
         253,191,111; ...
         255,127,0;   ...
         202,178,214; ...
         106,61,154; ...
         255,255,153; ...
         177,89,40]./256;

    r = x(rem(0:(n-1),size(x,1))+1,:);
end
